  $ $MERLIN single locate -look-for ml -position 6:6 \
  > -filename ./environment_on_open.ml < ./environment_on_open.ml
  {
    "class": "return",
    "value": {
      "file": "$TESTCASE_ROOT/environment_on_open.ml",
      "pos": {
        "line": 1,
        "col": 7
      }
    },
    "notifications": []
  }
