val ( = )   : int -> int -> bool
val ( <> )  : int -> int -> bool
val ( < )   : int -> int -> bool
val ( > )   : int -> int -> bool
val ( <= )  : int -> int -> bool
val ( >= )  : int -> int -> bool

val compare : int -> int -> int
