(**************************************************************************)
(*                                                                        *)
(*                                 OCaml                                  *)
(*                                                                        *)
(*             Sebastien Hinderer, projet Cambium, INRIA Paris            *)
(*                                                                        *)
(*   Copyright 2022 Institut National de Recherche en Informatique et     *)
(*     en Automatique.                                                    *)
(*                                                                        *)
(*   All rights reserved.  This file is distributed under the terms of    *)
(*   the GNU Lesser General Public License version 2.1, with the          *)
(*   special exception on linking described in the file LICENSE.          *)
(*                                                                        *)
(**************************************************************************)

(* Generate an .annot file from a .cmt file. *)

val gen_annot :
  string option ->
  sourcefile:string option ->
  use_summaries:bool -> Cmt_format.binary_annots ->
  unit

val iterator : scope:Location.t -> bool -> Tast_iterator.iterator

val binary_part : Tast_iterator.iterator -> Cmt_format.binary_part -> unit
