(* TODO or FIXME: If file is empty, Merlin just drops the errors
   (errors are attached to typed top-level definitions,
    if there are no definitions, errors cannot be attached!)

   So here is a dummy definition.
*)


let () = print_string "Hello world"
