
  $ $MERLIN single locate -look-for ml -position 7:17 -filename ./issue798.ml < ./issue798.ml
  {
    "class": "return",
    "value": {
      "file": "$TESTCASE_ROOT/issue798.ml",
      "pos": {
        "line": 6,
        "col": 8
      }
    },
    "notifications": []
  }
